// 008 u_c_text_file
// 22 jun 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_text_file;
  interface
    uses u_types_constants, u_c_basic_file, u_c_line, u_characters;

    type c_text_file= class(c_basic_file)
                        public
                          m_pt_buffer: t_oa_characters;
                          m_buffer_size, m_buffer_index: Integer;

                          m_has_line_feed: Boolean;

                          m_c_line: c_line;
                          m_from_to_end, m_from_to_begin: Integer;

                          constructor create_text_file(p_name, p_file_name: String); Virtual;

                          function f_load: Boolean; virtual;
                          function f_load_with_fat(p_factor: Integer): Boolean; virtual;
                          procedure assign_buffer(p_pt: Pointer; p_size: Integer);
                          function f_read_line: Boolean; virtual;
                          function f_end_of_text: boolean; virtual;

                          procedure list; virtual;
                          procedure dump_hex(p_begin, p_end: Integer);

                          function f_save: Boolean;
                          procedure save_in_u_directory_with_extension(p_name, p_extension: String);

                          function f_write(p_text: String): Boolean;
                          function f_write_line(p_text: String): Boolean;

                          function f_read_from_to(p_begin, p_end: Integer): String;
                          function f_write_from_to(p_text: String): Boolean;

                          function f_extract_string_start_length(p_start, p_length: Integer): String;
                          function f_extract_string_start_end(p_start, p_end: Integer): String;
                          procedure skip_alternate_return_line_feed(var pv_index: Integer);
                          procedure skip_end_of_line(var pv_index: Integer);

                          function f_ascii_signature: Integer;

                          destructor Destroy; Override;
                      end;

  implementation
    uses SysUtils, FileCtrl, u_display_hex_2,
      u_c_display, u_c_file_name;

    constructor c_text_file.create_text_file(p_name, p_file_name: String);
      begin
        Inherited create_basic_file(p_name, p_file_name);

        m_c_line:= c_line.create_basic_object(p_name+ '_line');
      end; // create_text_file

    function c_text_file.f_load: Boolean;
      begin
        if f_open
          then begin
              m_buffer_size:= f_file_size;

              SetLength(m_pt_buffer, m_buffer_size);
              if f_block_read(Pointer( @ m_pt_buffer[0] ), m_buffer_size)
                then begin
                    m_buffer_index:= 0;
                    Result:= True;
                  end
                else begin
                    display_bug('read error '+ m_c_file_name.m_complete_name);
                    Result:= False;
                  end;

              close_file;
            end
          else Result:= False;
      end; // f_load

    function c_text_file.f_load_with_fat(p_factor: Integer): Boolean;
      begin
        if f_open
          then begin
              m_buffer_size:= f_file_size;

              SetLength(m_pt_buffer, p_factor* m_buffer_size);
              if f_block_read(Pointer( @ m_pt_buffer[0] ), m_buffer_size)
                then begin
                    m_buffer_index:= 0;
                    Result:= True;
                  end
                else begin
                    display_bug('read error '+ m_c_file_name.m_complete_name);
                    Result:= False;
                  end;

              close_file;
            end
          else Result:= False;
      end; // f_load_with_fat

    procedure c_text_file.assign_buffer(p_pt: Pointer; p_size: Integer);
      begin
        m_buffer_size:= p_size;
        SetLength(m_pt_buffer, m_buffer_size);
        if m_buffer_size> 0
          then Move(p_pt^, m_pt_buffer[0], m_buffer_size);
        m_buffer_index:= 0;
      end; // assign_buffer

    function c_text_file.f_read_line: Boolean;
        // False if no more text
      var l_start_index: integer;
      begin
        l_start_index:= m_buffer_index;

        while (m_buffer_index< m_buffer_size) and not (m_pt_buffer[m_buffer_index] in [k_return, k_line_feed]) do
          m_buffer_index:= m_buffer_index+ 1;

        with m_c_line do
        begin
          m_buffer_index_of_start:= l_start_index;

          m_line_size:= m_buffer_index- l_start_index;
          if m_line_size< 0
            then display_bug_halt('f_read_line, size '+ IntToStr(m_buffer_index));

          if m_line_size= 0
            then begin
                m_the_line:= '';
                Result:= True;
              end
            else
              if (m_line_size>= 1) and (m_pt_buffer[l_start_index]= chr(26))
                then begin
                    // Destroy Ctrl-Z of MS DOS
                    m_line_size:= 0;
                    m_the_line:= '';
                    Result:= False;
                  end
                else begin
                    SetLength(m_the_line, m_line_size);
                    move(m_pt_buffer[l_start_index], m_the_line[1], m_line_size);
                    m_line_index:= 1;

                    Result:= True;
                  end;

          inc(m_line_line_number);
        end; // with m_c_line

        if (m_buffer_index< m_buffer_size) and (m_pt_buffer[m_buffer_index]= k_return)
          then inc(m_buffer_index, 1);

        if (m_buffer_index< m_buffer_size) and (m_pt_buffer[m_buffer_index]= k_line_feed)
          then begin
              inc(m_buffer_index, 1);
              m_has_line_feed:= true;
            end
          else m_has_line_feed:= false;
      end; // f_read_line

    function c_text_file.f_end_of_text: boolean;
      begin
        Result:= m_buffer_index>= m_buffer_size;
      end; // f_end_of_text

    procedure c_text_file.list;
      begin
        if f_load
          then begin
              display_line;
              display('list of '+ m_c_file_name.m_complete_name);
              display_line;

              while not f_end_of_text do
              begin
                f_read_line;
                with m_c_line do
                  display(Format('%5d >', [m_line_line_number])+ m_the_line+ '<');
              end; // while
            end
          else display('*** could not open '+ m_c_file_name.m_complete_name
              + ' '+ IntToStr(m_io_result));
      end; // list

    procedure c_text_file.dump_hex(p_begin, p_end: Integer);
      begin
        display(f_display_hex(Pointer(@ m_pt_buffer[p_begin]), p_end+ 1- p_begin, 16, k_hex_default));
      end; // dump_hex

    function c_text_file.f_save: Boolean;
      begin
        if f_open
          then begin
              seek_file(0);

              if f_block_write(Pointer(@ m_pt_buffer[0]), m_buffer_size)
                then Result:= True
                else begin
                    display_bug('write error '+ m_c_file_name.m_complete_name);
                    Result:= False;
                  end;

              close_file;
            end
          else Result:= False;
      end; // f_save

    procedure c_text_file.save_in_u_directory_with_extension(p_name, p_extension: String);
      var l_drive_and_path: String;
      begin
        with f_c_complete_file_name('name', p_name) do
        begin
          display_name_parts;
          m_extension:= p_extension;
          m_path:= m_path+ 'u\';
          compute_complete_name;
          l_drive_and_path:= f_drive_and_path;

          if not DirectoryExists(l_drive_and_path)
            then ForceDirectories(l_drive_and_path);

          display_name_parts;
          display('save into '+ m_complete_name);
          change_complete_name(m_complete_name);

          Free;
        end;

        if f_force_create
          then begin
              close_file;
              f_save;
            end;
      end; // save_in_u_directory_with_extension

    function c_text_file.f_write(p_text: String): Boolean;
      var l_length: Integer;
      begin
        l_length:= Length(p_text);
        if l_length= 0
          then Result:= True
          else Result:= f_block_write(Pointer(p_text), l_length);
      end; // f_write

    function c_text_file.f_write_line(p_text: String): boolean;
      var l_length: Integer;
      begin
        p_text:= p_text+ k_return+ k_line_feed;
        l_length:= Length(p_text);
        Result:= f_block_write(Pointer(p_text), l_length);
      end; // f_write_line

    function c_text_file.f_read_from_to(p_begin, p_end: Integer): String;
      var l_length: Integer;
      begin
        if p_begin> m_size
          then begin
              Result:= '';
              m_from_to_begin:= 0;
              m_from_to_end:= 0;
            end
          else begin
              seek_file(p_begin);
              if p_end> m_size
                then p_end:= m_size;
              l_length:= p_end- p_begin;
              SetLength(Result, l_length);
              f_block_read(Pointer(Result), l_length);

              m_from_to_begin:= p_begin;
              m_from_to_end:= f_file_position;
            end;
      end; // f_read_from_to

    function c_text_file.f_write_from_to(p_text: String): Boolean;
        // p_begin and p_end: the values used by the read_from_to
        // there must be NO other op in beteween
      var l_text_length: Integer;

      function f_put_in_old_location: Boolean;
        var l_from_to_length: Integer;
        begin
          l_from_to_length:= m_from_to_end- m_from_to_begin;

          display('  from_to_len '+ IntToStr(l_from_to_length)
            +  ' new '+ IntToStr(l_text_length));
          if (m_from_to_begin= 0)
              or (m_from_to_begin>= m_size)
              or (m_from_to_end>= m_size)
              or (l_from_to_length<= 0)
            then begin
                display('*** incoherent previous from-to values');
                Result:= False;
              end
            else
              if l_text_length> l_from_to_length
                then begin
                    display('greater: put at end');
                    Result:= False;
                  end
                else begin
                  seek_file(m_from_to_begin);
                  Result:= f_block_write(Pointer(p_text), l_text_length);

                  m_write_begin:= m_from_to_begin;
                  m_write_end:= f_file_position;

                  display('write inside old ' + IntToStr(m_write_begin)+ '..'+ IntToStr(m_write_end));
                end;
        end; // f_put_in_old_location

      begin // f_write_from_to
        l_text_length:= Length(p_text);

        if l_text_length= 0
          then Result:= True
          else begin
              if f_put_in_old_location
                then Result:= True
                else begin
                    display('size '+ IntToStr(m_size)+ ' '+ IntToStr(f_file_size));
                    seek_file(m_size);
                    Result:= f_block_write(Pointer(p_text), l_text_length);
                    display('write at end ' + IntToStr(m_write_begin)+ '..'+ IntToStr(m_write_end));
                  end;
            end;
      end; // f_write_from_to

    function c_text_file.f_extract_string_start_length(p_start, p_length: Integer): String;
      begin
        if p_length= 0
          then Result:= ''
          else begin
              if (p_length< 0) or (p_start+ p_length> m_buffer_size)
                then begin
                    stop('*** f_extract_string_start_length '
                        + m_c_file_name.m_complete_name+ ' '
                        + IntToStr(p_start)+ ' '+ IntToStr(p_length)+ ' '+ IntToStr(m_buffer_size));
                    Result:= '';
                  end
                else begin
                    SetLength(Result, p_length);
                    Move(m_pt_buffer[p_start], Result[1], p_length);
                  end;
            end;
      end; // f_extract_string_start_length

    function c_text_file.f_extract_string_start_end(p_start, p_end: Integer): String;
      begin
        Result:= f_extract_string_start_length(p_start, p_end+ 1- p_start)
      end; // f_extract_string_start_end

    procedure c_text_file.skip_alternate_return_line_feed(var pv_index: Integer);
        // has found either RET or LF
        // if other, skip it
      var l_return_or_line_feed: Char;
      begin
        if pv_index>= m_buffer_size
          then begin
              display('*** c_text_file.skip_alternate_return_line_feed '+ IntToStr(pv_index));
              Halt;
            end;
        l_return_or_line_feed:= m_pt_buffer[pv_index];

        // -- skip this one
        inc(pv_index);

        if   (pv_index< m_buffer_size)
            and
              (
                  (l_return_or_line_feed= k_return) and (m_pt_buffer[pv_index]= k_line_feed)
                or
                  (l_return_or_line_feed= k_line_feed) and (m_pt_buffer[pv_index]= k_return)
              )
          then inc(pv_index);
      end; // skip_alternate_return_line_feed

    procedure c_text_file.skip_end_of_line(var pv_index: Integer);
      begin
        while (pv_index< m_buffer_size)
            and (m_pt_buffer[pv_index]<> k_line_feed)
            and (m_pt_buffer[pv_index]<> k_return) do
          inc(pv_index);

        skip_alternate_return_line_feed(pv_index);
      end; // skip_end_of_line

    function c_text_file.f_ascii_signature: Integer;
      var l_index: Integer;
      begin
        Result:= 0;
        l_index:= 0;
        while l_index< m_buffer_size do
        begin
          if not (m_pt_buffer[l_index]
              in [k_return, k_line_feed, ' ', k_tabulation, '*', '{', '(', '}', ')'])
            then Inc(Result);
          Inc(l_index);
        end;
      end; // f_ascii_signature

    destructor c_text_file.Destroy;
      begin
        m_c_line.Free;

        Inherited Destroy;
      end; // Destroy

    begin // u_text_file
    end. // u_text_file

