// 002 u_c_http_client
// 31 oct 2006

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_http_client;
  interface
    uses ScktComp
        , u_c_basic_object
        , u_c_byte_buffer
        , u_c_url
        ;

    const // -- used as the basis for detecting reception buffer overflow,
          // --  (then doubles the capacity)
          // -- could use the TCP/IP max size MPU (1500...)
          k_tcp_ip_chunk= 1024;
          k_http_port= 80;

    type c_http_client= class; // Forward
         t_po_received_event= Procedure(p_c_http_client: c_http_client) of Object;

         c_http_client= class(c_basic_object)
                          m_c_client_socket: tClientSocket;
                          m_c_url: c_url;
                          m_c_reception_buffer: c_byte_buffer;
                          m_total_received_bytes: Integer;

                          m_on_received_event: t_po_received_event;

                          Constructor create_http_client(p_name: String);

                            procedure handle_socket_error(p_c_client_socket: tObject;
                                p_c_winsocket: TCustomWinSocket;
                                p_error_event: TErrorEvent; var pv_error_code: Integer);
                            procedure handle_after_connection(p_c_client_socket: tObject;
                                p_c_winsocket: TCustomWinSocket);
                            procedure handle_write(p_c_client_socket: tObject;
                                p_c_winsocket: TCustomWinSocket);
                            procedure handle_read(p_c_client_socket: tObject;
                                p_c_winsocket: TCustomWinSocket);
                            procedure disconnect;
                            procedure handle_after_disconnection(p_c_client_socket: tObject;
                                p_c_winsocket: TCustomWinSocket);

                          procedure connect;

                          procedure download_page(p_url: String;
                              p_po_received_event: t_po_received_event);

                          Destructor Destroy; Override;
                        end; // c_http_client

  implementation
    uses SysUtils, WinSock
        , u_types_constants, u_characters, u_c_display
        , u_winsock
        ;

    // -- c_http_client

    Constructor c_http_client.create_http_client(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        m_c_reception_buffer:= c_byte_buffer.create_byte_buffer('receive', 2* k_tcp_ip_chunk);

        m_c_client_socket:= tClientSocket.Create(Nil);
        with m_c_client_socket do
        begin
          Port:= k_http_port;

          OnError:= handle_socket_error;
          OnConnect:= handle_after_connection;
          OnWrite:= handle_write;
          OnRead:= handle_read;
          OnDisconnect:= handle_after_disconnection;
        end;
      end; // create_http_client

    procedure c_http_client.handle_socket_error(p_c_client_socket: tObject;
          p_c_winsocket: TCustomWinSocket;
          p_error_event: TErrorEvent; var pv_error_code: Integer);
      begin
        display('*** error '+ Format('%7d ', [pv_error_code])+ f_socket_error_message(pv_error_code));
        pv_error_code:= 0;
      end; // handle_socket_error

    procedure c_http_client.connect;
      // var l_in_address: tInAddr;
      begin
        display('+ c_http_client.connect');
        if m_c_client_socket.Active
          then m_c_client_socket.Active:= False;

        with m_c_client_socket do
        begin
          Host:= m_c_url.m_domain;
          Open;
        end;
      end; // connect

    procedure c_http_client.handle_after_connection(p_c_client_socket: tObject;
        p_c_winsocket: TCustomWinSocket);
      var l_page, l_host: String;
          l_get_request: String;
          l_result: Integer;
      begin
        display('= connected ');
        // -- now sent the HTTP request

        with m_c_url do
        begin
          l_page:= m_segment+ m_page+ m_extension;
          l_host:= m_domain;
        end; // with m_c_url

        l_get_request:= 'GET '+ l_page+ ' HTTP/1.0'+ k_new_line
            + 'Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, */*'+ k_new_line
            + 'User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)'+ k_new_line
            + 'Host: '+ l_host+ k_new_line
            + k_new_line;

        display('get '+ l_get_request+ '');    

        l_result:= p_c_winsocket.SendBuf(Pointer(l_get_request)^, Length(l_get_request));
      end; // handle_after_connection

    procedure c_http_client.handle_write(p_c_client_socket: tObject;
        p_c_winsocket: TCustomWinSocket);
      begin
        display('= Write ');
      end; // handle_write

    procedure c_http_client.handle_read(p_c_client_socket: tObject;
        p_c_winsocket: TCustomWinSocket);
      var l_bytes_received: Integer;
          l_remaining: Integer;
          l_pt_start_reception: Pointer;
      begin
        display('= receive ');

        // -- nothing tells us how much arrived. Grab as
        // --   much as possible. If no more was there,
        // --   Recv will return -1
        // -- And will even get several OnRead events, with
        // --   nothing in the reception buffer
        with m_c_reception_buffer do
        begin
          // -- add the received data to the current buffer
          repeat
            l_remaining:= m_buffer_size- m_write_index;

            // -- if not at least a tcp-ip chunk, increase the room
            if l_remaining< k_tcp_ip_chunk
              then begin
                  display('  double '+ IntToStr(l_remaining));
                  // -- reallocate
                  double_the_capacity;
                  l_remaining:= m_buffer_size- m_write_index;
                end;
            display('  remain '+ IntToStr(l_remaining));

            display('  write '+ Format('wr_ix %7d ', [m_write_index]));
            l_pt_start_reception:= @ m_oa_byte_buffer[m_write_index];
            // -- get at most k_tcp_ip_chunk
            l_bytes_received:= p_c_winsocket.ReceiveBuf(l_pt_start_reception^, k_tcp_ip_chunk);
            if l_bytes_received<= 0
              then begin
                  display('  -1 => no_more_for_now');
                  Break;
                end;

            if l_bytes_received> 0
              then begin
                  Inc(m_write_index, l_bytes_received);
                  Inc(m_total_received_bytes, l_bytes_received);
                end; // if l_packet_bytes

            display(Format('    received %7d tot %7d  wr_ix %7d  max_buf %7d bf_sz %7d',
                [l_bytes_received, m_total_received_bytes, m_write_index, Length(m_oa_byte_buffer), m_buffer_size]));
          until False;
        end; // with m_c_reception_buffer
      end; // handle_read

    procedure c_http_client.disconnect;
      begin
        display('+ c_http_client.disconnect');
        m_c_client_socket.Active:= False;
      end; // disconnect

    procedure c_http_client.handle_after_disconnection(p_c_client_socket: tObject;
        p_c_winsocket: TCustomWinSocket);
      begin
        display('= disconnected ');
        if Assigned(m_on_received_event)
          then m_on_received_event(Self);
      end; // handle_after_disconnection

    procedure c_http_client.download_page(p_url: String;
        p_po_received_event: t_po_received_event);
      begin
        display('> c_http_client.download_page '+ p_url);

        m_c_url.Free;
        m_c_url:= c_url.create_url(p_url);
        if m_c_url.f_decompose_url
          then begin
              display(m_c_url.f_display_url);

              m_on_received_event:= p_po_received_event;
              connect;
            end;
        display('< c_http_client.download_page ');
      end; // download_page

    Destructor c_http_client.Destroy;
      begin
        m_c_client_socket.Free;
        Inherited;
      end; // Destroy

    begin // u_c_http_client
    end. // u_c_http_client


