// 003 u_c_blog_item_list
// 06 nov 2006

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

// -- documentation: http://www.felix-jcolibri.com

(*$r+*)

unit u_c_blog_item_list;
  interface
    uses Classes
        , ComCtrls // tTreeView
        , u_c_basic_object
        ;

    type t_html_style= (e_unknown_html_style, e_html_bullet_style, e_html_table_style);
         c_blog_item= // one "blog_item"
                      Class(c_basic_object)
                        // -- m_name: the guid
                        m_title, m_url, m_publication_date, m_description: String;

                        Constructor create_blog_item(p_name,
                            p_title, p_url, p_description: String);
                        function f_display_blog_item: String;
                        function f_publication_datetime: tDateTime;
                        function f_c_self: c_blog_item;
                        Destructor Destroy; Override;
                      end; // c_blog_item

         c_blog_item_list= // "blog_item" list
                           Class(c_basic_object)
                             // -- the address to fetch the rss
                             m_blog_rss_url: String;

                             // -- the rss header attributes
                             m_title: String;

                             // -- the items of the rss
                             m_c_blog_item_list: tStringList;

                             Constructor create_blog_item_list(p_name: String);

                             function f_c_self: c_blog_item_list;
                             function f_display_blog: String;

                             function f_blog_item_count: Integer;
                             function f_c_blog_item(p_blog_item_index: Integer): c_blog_item;
                             function f_index_of(p_blog_item_name: String): Integer;
                             function f_c_find_by_blog_item(p_blog_item_name: String): c_blog_item;
                             procedure add_blog_item(p_blog_item_name: String; p_c_blog_item: c_blog_item);
                             function f_c_add_blog_item(p_blog_item_name,
                                 p_title, p_url, p_description: String): c_blog_item;
                             procedure display_blog_item_list;

                             procedure build_treeview(p_c_treeview: tTreeView);

                             function f_c_generate_html(p_html_style: t_html_style): tStringList;

                             Destructor Destroy; Override;
                           end; // c_blog_item_list

  implementation
    uses SysUtils, u_c_display
        , u_strings, u_characters
        , u_c_html_generator
        , u_html
        ;

    // -- c_blog_item

    Constructor c_blog_item.create_blog_item(p_name, p_title, p_url, p_description: String);
      begin
        Inherited create_basic_object(p_name);

        m_title:= p_title;
        m_url:= p_url;
        m_description:= p_description;
      end; // create_blog_item

    function c_blog_item.f_display_blog_item: String;
      begin
        Result:= Format('%-10s ', [m_name]);
      end; // f_display_blog_item

    function c_blog_item.f_c_self: c_blog_item;
      begin
        Result:= Self;
      end; // f_c_self

    function c_blog_item.f_publication_datetime: tDateTime;
        // -- |<pubDate>Mon, 23 Oct 2006 16:27:00
      var l_index: Integer;
          l_day_string, l_month_string, l_year_string: String;
      begin
        if m_publication_date= ''
          then Result:= 0
          else begin

            end;
      end; // f_publication_datetime

    Destructor c_blog_item.Destroy;
      begin
        InHerited;
      end; // Destroy

    // -- c_blog_item_list

    Constructor c_blog_item_list.create_blog_item_list(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        m_title:= p_name;

        m_c_blog_item_list:= tStringList.Create;
      end; // create_blog_item_line

    function c_blog_item_list.f_display_blog: String;
      begin
        Result:= Format('%-10s ', [m_name]);
      end; // f_display_blog

    function c_blog_item_list.f_c_self: c_blog_item_list;
      begin
        Result:= Self;
      end; // f_c_self

    // -- the item lists

    function c_blog_item_list.f_blog_item_count: Integer;
      begin
        Result:= m_c_blog_item_list.Count;
      end; // f_blog_item_count

    function c_blog_item_list.f_c_blog_item(p_blog_item_index: Integer): c_blog_item;
      begin
        Result:= c_blog_item(m_c_blog_item_list.Objects[p_blog_item_index]);
      end; //  f_c_blog_item

    function c_blog_item_list.f_index_of(p_blog_item_name: String): Integer;
      begin
        Result:= m_c_blog_item_list.IndexOf(p_blog_item_name);
      end; // f_index_of

    function c_blog_item_list.f_c_find_by_blog_item(p_blog_item_name: String): c_blog_item;
      var l_index_of: Integer;
      begin
        l_index_of:= f_index_of(p_blog_item_name);
        if l_index_of< 0
          then Result:= Nil
          else Result:= c_blog_item(m_c_blog_item_list.Objects[l_index_of]);
      end; // f_c_find_by_name

    procedure c_blog_item_list.add_blog_item(p_blog_item_name: String; p_c_blog_item: c_blog_item);
      begin
        m_c_blog_item_list.AddObject(p_blog_item_name, p_c_blog_item);
      end; // add_blog_item

    function c_blog_item_list.f_c_add_blog_item(p_blog_item_name,
        p_title, p_url, p_description: String): c_blog_item;
      begin
        Result:= c_blog_item.create_blog_item(p_blog_item_name, p_title, p_url, p_description);

        add_blog_item(p_blog_item_name, Result);
      end; // f_c_add_blog_item

    procedure c_blog_item_list.display_blog_item_list;
      var l_blog_item_index: Integer;
      begin
        display(m_name+ ' '+ IntToStr(f_blog_item_count));

        for l_blog_item_index:= 0 to f_blog_item_count- 1 do
          display(f_c_blog_item(l_blog_item_index).f_display_blog_item);
      end; // display_blog_item_list

    procedure c_blog_item_list.build_treeview(p_c_treeview: tTreeView);
      var l_c_root_tree_node: tTreeNode;
          l_blog_item_index: Integer;
      begin
        p_c_treeview.Items.BeginUpdate;

        // -- built the root node
        l_c_root_tree_node:= p_c_treeview.Items.AddChild(nil, 'ROOT '+ m_title);

        for l_blog_item_index:= 0 to f_blog_item_count- 1 do
          with f_c_blog_item(l_blog_item_index) do
          begin
             p_c_treeview.Items.AddChildObject(l_c_root_tree_node,
               m_title, f_c_self);
          end;

        p_c_treeview.Items.EndUpdate;
      end; // build_treeview

    function c_blog_item_list.f_c_generate_html(p_html_style: t_html_style): tStringList;
      var l_c_html_generator: c_html_generator;

      function f_format_text(p_text: String; p_column_max: Integer): String;
          // -- insert a return every p_column_max
        var l_index, l_column_size, l_previous_space, l_start_index: Integer;
            l_current_line: String;

        procedure flush_line;
          begin
            display(Format('%5d %5d %5d %5d', [l_start_index, l_previous_space, l_index, Length(p_text)]));
            if l_start_index> Length(p_text)
              then Exit;

            if l_previous_space= 0
              then begin
                  if l_index>= l_start_index
                    then Result:= Result+ f_extract_string(p_text, l_start_index, l_index)+ k_new_line;
                  l_start_index:= l_index+ 1;
                end
              else begin
                  if l_previous_space>= l_start_index
                    then Result:= Result+ f_extract_string(p_text, l_start_index, l_previous_space)+ k_new_line;
                  l_start_index:= l_previous_space+ 1;
                end;
            l_previous_space:= 0;
            l_column_size:= 0;
          end; // flush_line

        begin // f_format_text
          Result:= '';
          l_current_line:= '';
          l_column_size:= 0;
          l_start_index:= 1;
          l_previous_space:= 0;

          for l_index:= 1 to Length(p_text) do
          begin
            Inc(l_column_size);
            if p_text[l_index]= ' '
              then begin
                  if l_column_size> p_column_max
                    then begin
                        flush_line;
                      end
                    else l_previous_space:= l_index;
                end
              else begin
                  if l_column_size> p_column_max
                    then begin
                        flush_line;
                      end;
                end;
          end; // for l_index
          flush_line;
        end; // f_format_text

      procedure generate_bullet;
        var l_blog_item_index: Integer;
            l_text: String;
            l_header: String;
        begin
          with l_c_html_generator do
          begin
            add_bullet_list_start;

            for l_blog_item_index:= 0 to f_blog_item_count- 1 do
              with f_c_blog_item(l_blog_item_index) do
              begin
                l_header:= '<LI><FONT COLOR=#FF0000>'+ m_title
                    // + '[<A HREF="'+ m_url+ '">'+ m_url+ ' </A>';
                    + '[<A HREF="'+ m_url+ '">&nbsp;url&nbsp;</A>';
                if m_publication_date<> ''
                  then l_header := l_header+ ' '+ m_publication_date;
                l_header := l_header+ ']</FONT><BR>';

                add_line      (    l_header);

                l_text:= m_description;
                l_text:= f_html_remove_ampersand_string(l_text);

                add_line      (  l_text);
                add_line      (  '<HR>');
              end; // for l_blog_item_index, with f_c_blog_item

            add_bullet_list_end;
          end; // with l_c_html_generator
        end; // generate_bullet

      procedure generate_table;
        var l_blog_item_index: Integer;
            l_text: String;
        begin
          with l_c_html_generator do
          begin
            add_table_start(' border=1 width="400"');

            for l_blog_item_index:= 0 to f_blog_item_count- 1 do
              with f_c_blog_item(l_blog_item_index) do
              begin
                add_row_start ('');

                add_cell_start(  '');
                add_line      (    'title '+ m_title);
                add_cell_end  ();

                add_cell_start(  '');
                add_line      (    '<A HREF="'+ m_url+ '">ulr</A>');
                add_cell_end  ();

                add_cell_start(  '');

                l_text:= f_change_returns_in_spaces(m_description);
                l_text:= Trim(f_remove_double_spaces(l_text));
                display('rem_');
                l_text:= f_html_remove_ampersand_string(l_text);
                // -- insert returns all 60 columns
                l_text:= f_format_text(l_text, 55);

                display('_rem');
                add_line      (  l_text);
                add_cell_end  ();

                add_row_end();
              end; // for l_blog_item_index, with f_c_blog_item

            add_table_end;
          end; // with l_c_html_generator
        end; // generate_table

      begin // f_c_generate_html
        l_c_html_generator:= c_html_generator.create_html_generator('html_generator');

        with l_c_html_generator do
        begin
          add_line_indent  ('<HTML>');
          add_line_indent  ('<BODY>');
        end; // with l_c_html_generator

        case p_html_style of
          e_html_bullet_style : generate_bullet;
          e_html_table_style : generate_table;
        end; // case p_html_style

        with l_c_html_generator do
        begin
          unindent_add_line('</BODY>');
          unindent_add_line('</HTML>');

          // -- transfer the tStringList
          Result:= m_c_result_list;
          // -- do not detroy in when calls Free
          m_c_result_list:= Nil;

          Free;
        end; // with l_c_html_generator
      end; // f_c_generate_html

    Destructor c_blog_item_list.Destroy;
      var l_blog_item_index: Integer;
      begin
        for l_blog_item_index:= 0 to f_blog_item_count- 1 do
          f_c_blog_item(l_blog_item_index).Free;
        m_c_blog_item_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_blog
    end. // u_c_blog
