// 002 u_c_blog_reader
// 06 nov 2006

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_blog_reader;
  interface
    uses Classes, u_c_basic_object
      , u_c_http_client
      , u_c_url
      , u_c_xml_tree
      , u_c_blog_item_list
      ;

    type c_blog_reader= class; // forward
         t_po_blog_reader_event= Procedure(p_c_blog_reader: c_blog_reader) of Object;

         c_blog_reader= class(c_basic_object)
                          m_rss_path, m_rss_file_name: String;
                          // -- .HTML segment relative to m_rss_path
                          m_html_relative_segment: String;
                          m_rss_url: String;

                          // -- the content of the .RSS
                          m_c_rss_content: tStringList;

                          m_c_http_client: c_http_client;
                          m_on_downloaded_rss: t_po_blog_reader_event;

                          // -- the result of the .XML parser
                          m_c_xml_text: c_xml_text;

                          // -- the .XML tree in a more palatable (blog) presentation
                          m_c_blog_item_list: c_blog_item_list;

                          // -- the result of the .RSS -> .HTML transform
                          m_c_html_content: tStringList;

                          // -- stats
                          m_rss_size: Integer;
                          m_html_size: Integer;

                          Constructor create_blog_reader(p_name: String);

                            function f_rss_url_to_file_name(p_rss_url: String): String;
                          procedure download_rss(p_rss_url: String;
                              p_on_downloaded_rss: t_po_blog_reader_event);

                          procedure handle_received_rss_feed(p_c_http_client: c_http_client);

                          procedure load_rss_file(p_rss_path, p_rss_file_name: String);
                          procedure analyze_rss;
                          procedure _build_blog_item_list;

                            function f_rss_to_html_path: String;
                            function f_rss_to_html_file_name: String;
                          procedure rss_to_html(p_html_style: t_html_style);

                          Destructor Destroy; Override;
                        end; // c_blog_reader

  implementation
    uses SysUtils, u_c_display
      , u_characters, u_strings

      , u_c_xml_scanner
      , u_c_xml_parser_2
      ;

    // -- c_blog_reader

    Constructor c_blog_reader.create_blog_reader(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        // -- the .HTML segment relative to the .RSS path
        m_html_relative_segment:= '..\3_html_files\';
      end; // create_blog_reader

    function c_blog_reader.f_rss_url_to_file_name(p_rss_url: String): String;

      function f_specific_name(p_string: String): String;
        begin
          Result:= LowerCase(p_string);
          if (Result= 'feed') or (Result= 'rss') or (Result= 'blog') or (Result= '')
            then Result:= ''
            else Result:= '_'+ f_change_to_identifier_string(Result);
        end; // f_specific_name

      begin // f_rss_url_to_file_name
        // -- try to build a unique name
        with c_url.create_url(p_rss_url) do
        begin
          f_decompose_url;
          Result:= f_change_to_identifier_string(m_domain)
            + f_specific_name(m_segment)+ f_specific_name(m_page)+ '.txt';
          Free;
        end; // with c_url
      end; // f_rss_url_to_file_name

    procedure c_blog_reader.download_rss(p_rss_url: String;
        p_on_downloaded_rss: t_po_blog_reader_event);
      begin
        display('> download_rss');
        m_rss_url:= p_rss_url;
        // -- connect the user notification
        m_on_downloaded_rss:= p_on_downloaded_rss;
        // -- generate an rss_file_name from the url
        m_rss_file_name:= f_rss_url_to_file_name(p_rss_url);

        m_c_http_client.Free;
        m_c_http_client:= c_http_client.create_http_client('http_client');

        with m_c_http_client do
          download_page(p_rss_url, handle_received_rss_feed);
        display('< download_rss');
      end; // download_rss

    procedure c_blog_reader.handle_received_rss_feed(p_c_http_client: c_http_client);
        // -- here has received all the rss file
        // --  => analyze it and start IE
      var l_start_content_index: Integer;
      begin
        display('> handle_received_rss_feed');

        // -- save the file in order to analyze it
        with m_c_http_client, m_c_reception_buffer do
        begin
          l_start_content_index:= f_next_2_new_line_position(0)+ 4;
          m_rss_size:= m_write_index- l_start_content_index;

          // -- transfer to the tStringList
          m_c_rss_content.Free;
          m_c_rss_content:= tStringList.Create;
          m_c_rss_content.Text:= f_extract_string_start_end(l_start_content_index, m_write_index- 1);
        end; // with g_c_http_client, m_c_reception_buffer

        m_html_size:= 0;

        If Assigned(m_on_downloaded_rss)
          then m_on_downloaded_rss(Self);

        display('< handle_received_rss_feed');
      end; // handle_received_rss_feed

    procedure c_blog_reader.load_rss_file(p_rss_path, p_rss_file_name: String);
      begin
        display('> load_rss '+ p_rss_path+ p_rss_file_name);
        m_rss_path:= p_rss_path;
        m_rss_file_name:= p_rss_file_name;

        m_c_rss_content.Free;
        m_c_rss_content:= tStringList.Create;
        m_c_rss_content.LoadFromFile(m_rss_path+ m_rss_file_name);

        m_rss_size:= Length(m_c_rss_content.Text);
        display('< load_rss');
      end; // load_rss_file

    procedure c_blog_reader.analyze_rss;
        // -- load the .RSS and parse it
        // -- the result is in m_c_xml_text (an .XML tag/content tree)
      var l_c_xml_scanner: c_xml_scanner;
      begin
        display('> analyze_rss');
        m_c_xml_text.Free;
        m_c_xml_text:= c_xml_text.create_xml_text('xml_text');

        l_c_xml_scanner:= c_xml_scanner.create_xml_scanner('scanner',
            m_rss_path+ m_rss_file_name);

        with c_xml_parser.create_xml_parser('parser', l_c_xml_scanner) do
        begin
          // if Form1.trace_.Checked
          //  then m_trace_paser:= True;

          m_skip_tag_exclamation:= True;
          // m_trace_read_symbol:= true;
          parse_xml(m_c_xml_text);

          Free;
        end; // with c_xml_parser

        l_c_xml_scanner.Free;

        // -- create a more adapted structure
        _build_blog_item_list;

        display('< analyze_rss');
      end; // analyze_rss

    procedure c_blog_reader._build_blog_item_list;
      var l_item_count: Integer;
          l_c_current_blog_item: c_blog_item;

      procedure build_blog_item_list_recursive(p_level: Integer; p_c_xml_tag: c_xml_tag);
        var l_content_index: Integer;
        begin
          with p_c_xml_tag do
          begin
            if m_name= 'item'
              then begin
                  // add_row_start('');
                  Inc(l_item_count);

                  // -- could recurse here instead of counting items
                end;

            if m_name= 'title'
              then begin
                  if l_item_count= 1
                    then begin
                        m_c_blog_item_list.m_title:= Trim(f_c_get_strings_list.Text);
                      end
                    else begin
                        l_c_current_blog_item:=
                            m_c_blog_item_list.f_c_add_blog_item('item',
                                Trim(f_c_get_strings_list.Text), '', '');
                      end;

                end;
            if m_name= 'link'
              then begin
                  if l_item_count> 1
                    then l_c_current_blog_item.m_url:= Trim(f_c_get_strings_list.Text);
                end;
            if m_name= 'description'
              then begin
                  if l_item_count> 1
                    then l_c_current_blog_item.m_description:= Trim(f_c_get_strings_list.Text);
                end;
            if m_name= 'pubDate'
              then begin
                  if l_item_count> 1
                    then l_c_current_blog_item.m_publication_date:= Trim(f_c_get_strings_list.Text);
                end;

            with m_c_xml_tag_content_list do
              for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                with f_c_xml_tag_content(l_content_index) do
                  if f_c_self is c_xml_tag
                    then build_blog_item_list_recursive(p_level+ 1, f_c_xml_tag_content(l_content_index) as c_xml_tag)
                    else begin
                        // display('xxx_content');
                      end;
          end; // with p_c_xml_tag
        end; // build_blog_item_list_recursive

      begin // change_to_html_Click
        display('> build_blog_item_list');

        m_c_blog_item_list.Free;
        m_c_blog_item_list:= c_blog_item_list.create_blog_item_list('blog_items');

        build_blog_item_list_recursive(0, m_c_xml_text.m_c_xml_content);

        display('< build_blog_item_list');
      end; // _build_blog_item_list

    function c_blog_reader.f_rss_to_html_path: String;
        // -- hard codes to a sibbling folder
      begin
        Result:= m_rss_path+ m_html_relative_segment;
      end; // f_rss_to_html_path

    function c_blog_reader.f_rss_to_html_file_name: String;
      begin
        Result:= ChangeFileExt(m_rss_file_name, '.html')
      end; // f_rss_to_html_file_name

    procedure c_blog_reader.rss_to_html(p_html_style: t_html_style);
      begin
        m_c_html_content.Free;
        m_c_html_content:= m_c_blog_item_list.f_c_generate_html(p_html_style);

        m_html_size:= Length(m_c_html_content.Text);
      end; // rss_to_html

    Destructor c_blog_reader.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_blog_reader
    end. // u_c_blog_reader


