// 003 u_c_xml_parser_2
// 29 oct 2006

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_xml_parser_2;
  interface
    uses Classes
        , u_c_basic_object
        , u_c_xml_scanner
        , u_c_xml_tree
        ;

    type c_xml_parser= class(c_basic_object)
                         m_trace_paser: Boolean;
                         m_trace_read_symbol: Boolean;
                         m_c_xml_scanner_ref: c_xml_scanner;
                         m_skip_tag_exclamation: Boolean;

                         constructor create_xml_parser(p_name: String;
                             p_c_xml_scanner_ref: c_xml_scanner);
                         procedure parse_xml(p_c_xml_text: c_xml_text);
                         destructor Destroy; Override;
                       end; // c_xml_parser

  implementation
    uses SysUtils, u_types_constants, u_c_display, u_characters, u_strings;

    constructor c_xml_parser.create_xml_parser(p_name: String;
        p_c_xml_scanner_ref: c_xml_scanner);
      begin
        Inherited create_basic_object(p_name);
        m_c_xml_scanner_ref:= p_c_xml_scanner_ref;
      end; // create_xml_parser

    procedure c_xml_parser.parse_xml(p_c_xml_text: c_xml_text);
      var l_symbol_string: String;
          l_symbol_type: t_symbol_type;

      procedure display_trace(p_text: String);
        begin
          if m_trace_paser
            then display(p_text+ ' '+ l_symbol_string+ '|')
        end; // display_trace

      procedure display_parser_error(p_text: String);
        begin
          m_c_xml_scanner_ref.display_error(p_text);
          force_display(p_text);
          stop(p_text);
          display_bug_halt('c_xml_parser_2_parse_error');
        end; // display_parser_error

      procedure check(p_symbol_type: t_symbol_type);
        begin
          with m_c_xml_scanner_ref do
            if m_symbol_type<> p_symbol_type
              then display_parser_error(f_symbol_name(p_symbol_type)+ ' expected, but '
                  + f_symbol_name(l_symbol_type));
        end; // check

      var l_stopped_display: Boolean;

      procedure read_symbol;
        begin
          with m_c_xml_scanner_ref do
          begin
            get_symbol;

            repeat
              // display('sy ='+ f_symbol_name(m_symbol_type)+ ' '+ m_symbol_string);

              if m_symbol_type= e_comment_start_symbol
                then begin
                    repeat
                      get_symbol;
                    until m_symbol_type= e_comment_end_symbol;
                    get_symbol;
                  end
                else
                  if m_symbol_type= e_opening_comment_exclamation_symbol
                    then begin
                        repeat
                          get_symbol;
                        until m_symbol_type= e_closing_comment_exclamation_symbol;

                        get_symbol;
                      end
                    else
                      if m_skip_tag_exclamation and (m_symbol_type= e_opening_tag_exclamation_symbol)
                        then begin
                            repeat
                              get_symbol;
                            until m_symbol_type= e_closing_tag_symbol;

                            get_symbol;
                          end;
            until (m_symbol_type= e_end_symbol)
                or (
                      (m_symbol_type<> e_comment_start_symbol)
                    and
                      (m_symbol_type<> e_opening_comment_exclamation_symbol)
                    and
                      not ( m_skip_tag_exclamation and (m_symbol_type= e_opening_tag_exclamation_symbol))
                   );

            l_symbol_type:= m_symbol_type;
            l_symbol_string:= m_symbol_string;

            if m_trace_read_symbol
              then display('read_symbol '+ f_symbol_name(m_symbol_type));
          end; // with m_c_xml_scanner_ref

          if False // not l_stopped_display and (m_c_xml_scanner_ref.m_buffer_index> 200)
            then begin
                push_stop_display(1);
                l_stopped_display:= True;
              end;

        end; // read_symbol

      procedure read_text_symbol;
        begin
          with m_c_xml_scanner_ref do
          begin
            get_text_symbol;

            l_symbol_type:= m_symbol_type;
            l_symbol_string:= m_symbol_string;
          end; // with m_c_xml_scanner_ref
        end; // read_text_symbol

      function f_parse_tag_name: String;
        BEGIN
          display_trace('> tag_name');
          check(e_IDENTIFIER_symbol);

          Result:= '';

          Result:= Result+ l_symbol_string;
          read_symbol;

          while l_symbol_type in [e_colon_symbol, e_point_symbol] do
          BEGIN
            Result:= Result+ l_symbol_string;
            read_symbol;
            check(e_IDENTIFIER_symbol);
            Result:= Result+ l_symbol_string;
            read_symbol;
          END; // IF [

          display_trace('< tag_name %'+ Result+ '%');
        END; // f_parse_tag_name

      PROCEDURE parse_tag_attributes(p_c_xml_tag: c_xml_tag);
        var l_key: string;
            l_display: String;
        BEGIN
          display_trace('> tag_attributes');

          l_display:= '';

          // -- the key="value" part
          WHILE l_symbol_type= e_IDENTIFIER_symbol DO
          BEGIN
            // display('key '+ l_symbol_string);
            l_key:= l_symbol_string;
            l_display:= l_display+ l_key;
            read_symbol;

            // -- if key is "xxx:yyy= ..."
            while l_symbol_type in [e_colon_symbol] do
            BEGIN
              l_display:= l_display+ l_symbol_string;
              read_symbol;
              check(e_IDENTIFIER_symbol);
              l_display:= l_display+ l_symbol_string;
              read_symbol;
            END; // IF [

            check(e_equal_symbol);
            read_symbol;

            check(e_STRING_symbol);
            l_display:= l_display+ '='+ l_symbol_string+ ' | ';
            // display('string '+ l_symbol_string);
            p_c_xml_tag.add_attribute(l_key, l_symbol_string);
            read_symbol;
          END; // WHILE {

          display_trace('< tag_attributes '+ l_display);
        end; // parse_tag_attributes

      function f_c_parse_tag_name_and_attributes: c_xml_tag;
        BEGIN
          display_trace('> tag_name_and_attributes');

          Result:= c_xml_tag.create_xml_tag(f_parse_tag_name);
          parse_tag_attributes(Result);

          display_trace('< tag_name_and_attributes %'+ Result.f_display_xml_name_and_attributes+ '%');
        END; // f_c_parse_tag_name_and_attributes

      procedure parse_comment;
        begin
          display_trace('> parse_comment');

          check(e_comment_start_symbol);
          read_symbol;

          p_c_xml_text.m_c_xml_comment:= f_c_parse_tag_name_and_attributes;

          check(e_comment_end_symbol);
          read_symbol;

          display_trace('< parse_comment');
        end; // parse_comment

      function f_c_parse_tag_content_recursive: c_xml_tag;
        var l_c_xml_sub_tag: c_xml_tag;
            l_c_xml_string: c_xml_string;
        BEGIN
          display_trace('> tag_content');

          check(e_opening_tag_symbol);
          read_symbol;

          Result:= f_c_parse_tag_name_and_attributes;

          IF l_symbol_type= e_closing_tag_symbol
            THEN BEGIN
                // -- skip ">"
                // read_symbol;
                read_text_symbol;

                // -- await "</"
                while l_symbol_type<> e_opening_anti_tag_symbol do
                begin
                  IF l_symbol_type= e_opening_tag_symbol
                    THEN begin
                        l_c_xml_sub_tag:= f_c_parse_tag_content_recursive;
                        Result.m_c_xml_tag_content_list.add_xml_string(l_c_xml_sub_tag);
                      end
                    else begin
                        // -- any content
                        if not f_contains_only(l_symbol_string, [' ', k_tabulation, k_return, k_line_feed])
                          then begin
                              l_c_xml_string:= c_xml_string.create_xml_string(l_symbol_string);
                              Result.m_c_xml_tag_content_list.add_xml_string(l_c_xml_string);
                            end;

                        read_symbol;
                      end;
                end; // while

                // -- skip "</"
                check(e_opening_anti_tag_symbol);
                read_symbol;

                f_parse_tag_name;
                check(e_closing_tag_symbol);

                read_symbol;
              END // 2
            ELSE // e_1
              IF l_symbol_type= e_closing_anti_tag_symbol
                THEN
                  read_symbol
                ELSE display_parser_error('>, />');

          display_trace('< tag_content');
        END; // f_c_parse_tag_content_recursive

      begin  // parse_xml
        display('> parse_xml');

        l_symbol_string:= '';
        l_symbol_type:= e_unknown_symbol;
        l_stopped_display:= False;

        read_symbol;

        // parse_comment;
        p_c_xml_text.m_c_xml_content:= f_c_parse_tag_content_recursive;

        display('< parse_xml');
      end; // parse_xml

    destructor c_xml_parser.Destroy;
      begin
        Inherited Destroy;
      end; // Destroy

    begin // u_xml_1
    end. // u_xml_1


