// 003 u_c_xml_scanner
// 29 oct 2006

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_xml_scanner;
  interface
    uses Classes,
        u_c_text_file
        ;

    type t_symbol_type= (e_unknown_symbol,

                      e_string_symbol,
                      e_integer_symbol,
                      e_identifier_symbol,
                      e_free_text_symbol,

                      e_text_symbol,

                      e_opening_tag_symbol, // <
                      e_closing_tag_symbol, // >
                      e_colon_symbol, // :
                      e_point_symbol,
                      e_equal_symbol,

                      e_opening_anti_tag_symbol, // </
                      e_closing_anti_tag_symbol, // />
                      e_comment_start_symbol, // <?
                      e_comment_end_symbol, // ?>

                      e_opening_tag_exclamation_symbol, // <!   ex: <!DOCUMENT xxx>
                      e_opening_comment_exclamation_symbol, // <!--
                      e_closing_comment_exclamation_symbol, // -->

                      e_end_symbol
                      );

         c_xml_scanner= class(c_text_file)
                          m_file_name: String;
                          m_trace_xml: Boolean;
                          m_symbol_max: Integer;
                          // m_display_symbol_read: Boolean;

                          m_blank_string: String;
                          m_symbol_string: String;
                          m_comment_string: String;
                          m_symbol_type: t_symbol_type;

                          m_c_unknown_symbol_list: tStringList;

                          constructor create_xml_scanner(p_name,
                              p_file_name: String);

                          procedure display_error(p_text: String);
                          procedure get_symbol;
                          procedure get_text_symbol;
                          procedure get_non_comment_symbol;

                          procedure test_scanner(p_display: Boolean);
                          procedure test_scanner_max(p_display: Boolean; p_symbol_max: Integer);
                          procedure test_non_comment_scanner_max(p_display: Boolean; p_symbol_max: Integer);

                          destructor Destroy; Override;
                        end; // c_xml_scanner

    const k_comment_symbol= [e_comment_start_symbol, e_opening_comment_exclamation_symbol];

    function f_symbol_name(p_symbol_type: t_symbol_type): String;

  implementation
    uses SysUtils, Math, TypInfo
        , u_types_constants, u_c_display
        , u_characters, u_strings, u_stringlist
        ;

    function f_symbol_name(p_symbol_type: t_symbol_type): String;
      begin
        Result:= GetEnumName(TypeInfo(t_symbol_type), Integer(p_symbol_type));
        Delete(Result, 1, 2);
        Delete(Result, Length(Result)- Length('_symbol')+ 1, Length('_symbol'));
      end; // f_symbol_type

    constructor c_xml_scanner.create_xml_scanner(p_name,
        p_file_name: String);
      begin
        Inherited create_text_file(p_name, p_file_name);
        m_file_name:= p_file_name;

        if f_load
          then begin
              // display('load '+ IntToStr(m_buffer_size)+ ' '+ IntToStr(m_buffer_index)
              //  + ' '+ m_pt_buffer[m_buffer_index]);
              dump_hex(0, 7);
              m_buffer_index:= 0;

              if (m_buffer_size> 3)
                  and (m_pt_buffer[0]= chr($EF))
                  and (m_pt_buffer[1]= chr($BB))
                  and (m_pt_buffer[2]= chr($BF))
                then begin
                    display('UNICODE');
                    m_buffer_index:= 3;
                  end;
            end // load ok
          else display('*** could not open '+ m_c_file_name.m_complete_name
              + ' '+ IntToStr(m_io_result));

        // stop('');
        m_c_unknown_symbol_list:= tStringList.Create;
      end; // create_xml_scanner

    procedure c_xml_scanner.display_error(p_text: String);
      var l_index: Integer;
          l_text: String;
      begin
        l_index:= Max(0, m_buffer_index- 100);
        l_text:= f_extract_string_start_end(l_index, m_buffer_index- 1);
        display_line;
        display('==============>>>');
        display(l_text+ '|');
        display('');
        l_index:= Min(m_buffer_index+ 100, m_buffer_size- 1);
        l_text:= f_extract_string_start_end(m_buffer_index, l_index);
        display('|'+ l_text);
        display('==============<<<');
        display('ERROR '+ p_text);
      end; // display_error

    procedure c_xml_scanner.get_symbol;
      const k_blanks= [' ', chr(9), k_return, k_line_feed];
            k_identifier_characters= ['a'..'z'];
(*
            k_returns= [k_return, k_line_feed];

            k_tag_end= [' ', '>', chr(13)];
            k_other_end= k_returns+ ['<'];
*)
      procedure trace_xml(p_text: String);
        begin
          if m_trace_xml
            then display(p_text+ ' '+ f_symbol_name(m_symbol_type));
        end; // trace_xml

      procedure skip_blanks;
        var l_start_index: Integer;
        begin
          l_start_index:= m_buffer_index;
          while (m_buffer_index< m_buffer_size)
              and (m_pt_buffer[m_buffer_index] in k_blanks) do
          begin
            if (m_buffer_index< m_buffer_size)
                and (m_pt_buffer[m_buffer_index]= k_return)
              then begin
                  inc(m_buffer_index);

                  if (m_buffer_index< m_buffer_size)
                      and (m_pt_buffer[m_buffer_index]= k_line_feed)
                    then inc(m_buffer_index);
                end
              else Inc(m_buffer_index);
          end; // while blanks

          m_blank_string:= f_extract_string_start_end(l_start_index, m_buffer_index- 1);
        end; // skip_blanks

      procedure get_string_symbol(p_delimiter: Char);
          // -- the string without the quotes
        var l_start_index: Integer;
        begin
          // l_litteral:= '"';
          // inc(m_buffer_index);

          // -- skip the delimiter
          l_start_index:= m_buffer_index+ 1;
          repeat
            inc(m_buffer_index);
          until (m_buffer_index>= m_buffer_size)
              or (m_pt_buffer[m_buffer_index]= p_delimiter);

          m_symbol_string:= f_extract_string_start_end(l_start_index, m_buffer_index- 1);
          m_symbol_type:= e_string_symbol;

          // -- skip closing "
          inc(m_buffer_index);
        end; // get_string_symbol

      procedure get_number;
        var l_start_index: Integer;
            l_number: String;
        begin
          l_start_index:= m_buffer_index;

          if m_pt_buffer[m_buffer_index]= '-'
            then inc(m_buffer_index);
          while (m_buffer_index< m_buffer_size)
              and (m_pt_buffer[m_buffer_index] in k_digits) do
            inc(m_buffer_index);

          m_symbol_string:= f_extract_string_start_end(l_start_index, m_buffer_index- 1);
          m_symbol_type:= e_integer_symbol;
        end; // get_number

      procedure get_identifier;
        var l_start_index: Integer;
        begin
          l_start_index:= m_buffer_index;

          while (m_buffer_index< m_buffer_size)
              and (m_pt_buffer[m_buffer_index] in k_letters+ ['_', '0'..'9']) do
            inc(m_buffer_index);

          m_symbol_string:= f_extract_string_start_end(l_start_index, m_buffer_index- 1);
          m_symbol_type:= e_identifier_symbol;
        end; // get_identifier

      procedure get_operator_symbol(p_symbol_type: t_symbol_type);
        begin
          m_symbol_type:= p_symbol_type;
          m_symbol_string:= m_pt_buffer[m_buffer_index];
          Inc(m_buffer_index);
        end; // f_get_operator_symbol

      procedure get_2_operator_symbol(p_symbol_type: t_symbol_type);
        begin
          m_symbol_type:= p_symbol_type;
          m_symbol_string:= m_pt_buffer[m_buffer_index]+ m_pt_buffer[m_buffer_index+ 1];
          Inc(m_buffer_index, 2);
        end; // f_get_operator_symbol

      procedure get_unknown_symbol;
        begin
          m_symbol_type:= e_unknown_symbol;
          m_symbol_string:= m_pt_buffer[m_buffer_index];
          add_unique_name(m_c_unknown_symbol_list, m_symbol_string);
          Inc(m_buffer_index);
        end; // get_unknown_symbol

      begin // get_symbol
        if m_trace_xml
          then begin
(*
              if m_buffer_index< m_buffer_size
                then display(Format('%5d  ', [m_buffer_index])+ m_pt_buffer[m_buffer_index]);
*)
            end;

        m_symbol_type:= e_unknown_symbol;
        m_symbol_string:= '';

        skip_blanks;

        if m_buffer_index>= m_buffer_size
          then begin
              m_symbol_type:= e_end_symbol;
              exit;
            end;

        case m_pt_buffer[m_buffer_index] of
          '<' : if (m_buffer_index+ 1< m_buffer_size)
                  and (m_pt_buffer[m_buffer_index+ 1]= '/')
                    then get_2_operator_symbol(e_opening_anti_tag_symbol)
                    else
                      if (m_buffer_index+ 1< m_buffer_size)
                          and (m_pt_buffer[m_buffer_index+ 1]= '?')
                        then get_2_operator_symbol(e_comment_start_symbol)
                        else
                          if (m_buffer_index+ 3< m_buffer_size)
                              and (m_pt_buffer[m_buffer_index+ 1]= '!')
                              and (m_pt_buffer[m_buffer_index+ 2]= '-')
                              and (m_pt_buffer[m_buffer_index+ 3]= '-')
                            then begin
                                m_symbol_string:= '<!--';
                                m_symbol_type:= e_opening_comment_exclamation_symbol;
                                Inc(m_buffer_index, 4);
                              end
                            else
                              if (m_buffer_index+ 1< m_buffer_size)
                                  and (m_pt_buffer[m_buffer_index+ 1]= '!')
                                then get_2_operator_symbol(e_opening_tag_exclamation_symbol)
                                else get_operator_symbol(e_opening_tag_symbol);

          '>' : if (m_buffer_index> 2)
                    and (m_pt_buffer[m_buffer_index- 1]= '-')
                    and (m_pt_buffer[m_buffer_index- 2]= '-')
                  then begin
                      m_symbol_string:= '-->';
                      m_symbol_type:= e_closing_comment_exclamation_symbol;
                      Inc(m_buffer_index, 1);
                    end
                  else get_operator_symbol(e_closing_tag_symbol);
          '/' : if (m_buffer_index+ 1< m_buffer_size)
                  and (m_pt_buffer[m_buffer_index+ 1]= '>')
                    then get_2_operator_symbol(e_closing_anti_tag_symbol)
                    else get_operator_symbol(e_closing_tag_symbol);
          '?' : if (m_buffer_index+ 1< m_buffer_size)
                  and (m_pt_buffer[m_buffer_index+ 1]= '>')
                    then get_2_operator_symbol(e_comment_end_symbol)
                    else get_unknown_symbol;
          '=' : get_operator_symbol(e_equal_symbol);
          '.' : get_operator_symbol(e_point_symbol);
          ':' : get_operator_symbol(e_colon_symbol);
          '"', '''' : get_string_symbol(m_pt_buffer[m_buffer_index]);
          'a'..'z', 'A'..'Z' : get_identifier;

          else
            get_unknown_symbol
        end; // case

        if m_trace_xml
          then display(Format('%5d >', [m_buffer_index])+ m_symbol_string+ '< '+ f_symbol_name(m_symbol_type));
      end; // get_symbol

    procedure c_xml_scanner.get_non_comment_symbol;
      var l_blank_string: String;
      begin
        get_symbol;
        if m_symbol_type= e_comment_start_symbol
          then begin
              l_blank_string:= m_blank_string+ m_symbol_string;
              repeat
                get_symbol;
                if m_symbol_type= e_comment_end_symbol
                  then begin
                      l_blank_string:= m_blank_string+ m_symbol_string;
                      Break;
                    end
                  else l_blank_string:= m_blank_string+ m_symbol_string;
              until False or (m_symbol_type= e_end_symbol);

              get_symbol;
              m_blank_string:= l_blank_string+ m_blank_string;
            end
          else
            if m_symbol_type= e_opening_comment_exclamation_symbol
              then begin
                  l_blank_string:= m_blank_string+ m_symbol_string;
                  repeat
                    get_symbol;
                    if m_symbol_type= e_closing_comment_exclamation_symbol
                      then begin
                          l_blank_string:= m_blank_string+ m_symbol_string;
                          Break;
                        end
                      else l_blank_string:= m_blank_string+ m_symbol_string;
                  until False or (m_symbol_type= e_end_symbol);

                  get_symbol;
                  m_blank_string:= l_blank_string+ m_blank_string;
                end
              else ; // Break;
        // until not (m_symbol_type in k_comment_symbol);
      end; // get_non_comment_symbol

    procedure c_xml_scanner.get_text_symbol;
        // -- get anything between <xxx> and </xxx>
      var l_start_index: Integer;
      begin
        m_symbol_type:= e_unknown_symbol;
        m_symbol_string:= '';

        l_start_index:= m_buffer_index;

        while (m_buffer_index< m_buffer_size)
            and (m_pt_buffer[m_buffer_index]<> '<') do
          inc(m_buffer_index);

        m_symbol_string:= f_extract_string_start_end(l_start_index, m_buffer_index- 1);
        m_symbol_type:= e_free_text_symbol;
      end; // get_text_symbol

    procedure c_xml_scanner.test_scanner(p_display: Boolean);
      begin
        m_c_unknown_symbol_list.Clear;

        repeat
          get_symbol;
          if p_display
            then display(' |'+ m_symbol_string+ '| '+ f_symbol_name(m_symbol_type));
          if (m_symbol_max> 0) and (m_buffer_index> m_symbol_max)
            then Break;
        until m_symbol_type= e_end_symbol;
      end; // test_scanner

    procedure c_xml_scanner.test_scanner_max(p_display: Boolean; p_symbol_max: Integer);
      begin
        m_c_unknown_symbol_list.Clear;

        repeat
          get_symbol;
          Dec(p_symbol_max);
          if p_display
            then display(' |'+ m_symbol_string+ '| '+ f_symbol_name(m_symbol_type));
          if (m_symbol_max> 0) and (m_buffer_index> m_symbol_max)
            then Break;
        until (m_symbol_type= e_end_symbol) or (p_symbol_max<= 0);
      end; // test_scanner_max

    procedure c_xml_scanner.test_non_comment_scanner_max(p_display: Boolean; p_symbol_max: Integer);
      begin
        m_c_unknown_symbol_list.Clear;

        repeat
          get_non_comment_symbol;
          Dec(p_symbol_max);
          if p_display
            then display(' |'+ m_symbol_string+ '| '+ f_symbol_name(m_symbol_type));
          if (m_symbol_max> 0) and (m_buffer_index> m_symbol_max)
            then Break;
        until (m_symbol_type= e_end_symbol) or (p_symbol_max<= 0);
      end; // test_non_comment_scanner_max

    destructor c_xml_scanner.Destroy;
      begin
        m_c_unknown_symbol_list.Free;

        Inherited Destroy;
      end; // Destroy

    begin // u_xml_1
    end. // u_xml_1


